#include <string>
#include <tuple>

#include <signal.h>

#include <core/dbus/bus.h>
#include <core/dbus/object.h>
#include <core/dbus/service.h>

#include <core/dbus/asio/executor.h>

namespace org {
namespace gnome {
namespace SettingsDaemon {

static const std::chrono::milliseconds t{10000};

struct MediaKeys {
  static const std::string& name() { static const std::string s{"org.gnome.SettingsDaemon.MediaKeys"}; return s; }
  struct GrabMediaPlayerKeys {
    typedef MediaKeys Interface;
    static const std::string& name() { static const std::string s{"GrabMediaPlayerKeys"}; return s; }
    static const std::chrono::milliseconds& default_timeout() { return t; }
    typedef std::tuple<std::string,std::uint32_t> ArgumentType;
    typedef void ResultType;
  };
  struct ReleaseMediaPlayerKeys {
    typedef MediaKeys Interface;
    static const std::string& name() { static const std::string s{"ReleaseMediaPlayerKeys"}; return s; }
    static const std::chrono::milliseconds& default_timeout() { return t; }
    typedef std::string ArgumentType;
    typedef void ResultType;
  };
  struct Signals {
    struct MediaPlayerKeyPressed {
      typedef MediaKeys Interface;
      static const std::string& name() { static const std::string s{"MediaPlayerKeyPressed"}; return s; }
      typedef std::tuple<std::string,std::string> ArgumentType;
    };
  };
};

}  // namespace SettingsDaemon
}  // namespace gnome
}  // namespace org

int main(int, char**) {
  auto bus = std::make_shared<::core::dbus::Bus>(::core::dbus::WellKnownBus::session);
  bus->install_executor(::core::dbus::asio::make_executor(bus));
  std::thread t{std::bind(&::core::dbus::Bus::run, bus)};

  auto service = ::core::dbus::Service::use_service(bus, "org.gnome.SettingsDaemon");
  auto object = service->object_for_path(::core::dbus::types::ObjectPath("/org/gnome/SettingsDaemon/MediaKeys"));

  std::string app{"com.foo.bar"};
  auto arg = std::make_tuple(app, 0u);
  object->invoke_method_synchronously<org::gnome::SettingsDaemon::MediaKeys::GrabMediaPlayerKeys,
                                      void,
                                      org::gnome::SettingsDaemon::MediaKeys::GrabMediaPlayerKeys::ArgumentType>(arg);

  auto device_added_signal = object->get_signal<org::gnome::SettingsDaemon::MediaKeys::Signals::MediaPlayerKeyPressed>();
  device_added_signal->connect([](const org::gnome::SettingsDaemon::MediaKeys::Signals::MediaPlayerKeyPressed::ArgumentType &arg) {
      std::string application;
      std::string key;
      std::tie(application, key) = arg;
      std::cout << application << " " << key << std::endl;
    });

  sigset_t signal_set;
  sigemptyset(&signal_set);
  sigaddset(&signal_set, SIGINT);
  int signal;
  sigwait(&signal_set, &signal);

  object->invoke_method_synchronously<org::gnome::SettingsDaemon::MediaKeys::ReleaseMediaPlayerKeys,
                                      void,
                                      org::gnome::SettingsDaemon::MediaKeys::ReleaseMediaPlayerKeys::ArgumentType>(app);

  bus->stop();
  if (t.joinable())
    t.join();
}
